searchNodes=[{"doc":"An interface to define and leverage runtime resolution. Groups Mappings are defined under group/0s. Hook defaults the group to be the calling process making process isolation of mappings frictionless. Any term can also be supplied as the group. Putting :foo under :bar in the :red group, than fetching it from the :blue group will fail. Resolution Starting with the calling process: does the process define a mapping for the term 1.1. the term has been resolved does the process define fallbacks 2.1. for each fallback in order: goto step 1 does the process have ancestors 3.1. for each ancestor in order: goto step 1 the term has not been resolved Performance In general, the thing to focus on with respect to performance is whether the function call is getting serialized through a GenServer or not. You will find that only calls that result in a write get serialized and those that only read do not. Functions that get serialized through a GenServer: callback/3,4 fallback/1,2 put/2,3 resolve_callback/3 Functions that do not get serialized through a GenServer: assert/0,1 callbacks/0,1 fetch!/1,2 fetch/1,2 get_all/0,1 hook/1 This means that a &quot;hot path&quot; that is resolving a term via these functions should remain performant outside of extreme cases. ``","ref":"Hook.html","title":"Hook","type":"behaviour"},{"doc":"Asserts that all non-infinity callbacks defined via Hook.callback/4 have been consumed. Returns :ok or raises.","ref":"Hook.html#assert/1","title":"Hook.assert/1","type":"function"},{"doc":"","ref":"Hook.html#c:assert/1","title":"Hook.assert/1","type":"callback"},{"doc":"Defines a callback that can be consumed count times. Ensures a module value is put under the key module for the calling process.","ref":"Hook.html#callback/4","title":"Hook.callback/4","type":"function"},{"doc":"","ref":"Hook.html#c:callback/4","title":"Hook.callback/4","type":"callback"},{"doc":"Return group&#39;s callbacks.","ref":"Hook.html#callbacks/1","title":"Hook.callbacks/1","type":"function"},{"doc":"","ref":"Hook.html#c:callbacks/1","title":"Hook.callbacks/1","type":"callback"},{"doc":"Prepend group to the calling process&#39; fallbacks.","ref":"Hook.html#fallback/2","title":"Hook.fallback/2","type":"function"},{"doc":"","ref":"Hook.html#c:fallback/2","title":"Hook.fallback/2","type":"callback"},{"doc":"Fetches the value for key for group.","ref":"Hook.html#fetch/2","title":"Hook.fetch/2","type":"function"},{"doc":"","ref":"Hook.html#c:fetch/2","title":"Hook.fetch/2","type":"callback"},{"doc":"See Hook.fetch/2.","ref":"Hook.html#fetch!/2","title":"Hook.fetch!/2","type":"function"},{"doc":"","ref":"Hook.html#c:fetch!/2","title":"Hook.fetch!/2","type":"callback"},{"doc":"Gets all values for group.","ref":"Hook.html#get_all/1","title":"Hook.get_all/1","type":"function"},{"doc":"","ref":"Hook.html#c:get_all/1","title":"Hook.get_all/1","type":"callback"},{"doc":"A macro that compiles into term itself or a Hook.fetch(term) call. By default the fetch call will only be used when the Mix environment is not :prod. This logic can be overidden by the :should_hook configuration: config :hook, should_hook: {SomeModule, :some_function} :should_hook will be called with a single argument that is a term being processed by the hook/1 macro and must return a boolean(). This evaluation will happen at compile time.","ref":"Hook.html#hook/1","title":"Hook.hook/1","type":"macro"},{"doc":"Puts value under key for group.","ref":"Hook.html#put/3","title":"Hook.put/3","type":"function"},{"doc":"","ref":"Hook.html#c:put/3","title":"Hook.put/3","type":"callback"},{"doc":"Resolves a callback for the calling process and executes it with args.","ref":"Hook.html#resolve_callback/3","title":"Hook.resolve_callback/3","type":"function"},{"doc":"","ref":"Hook.html#c:resolve_callback/3","title":"Hook.resolve_callback/3","type":"callback"},{"doc":"","ref":"Hook.html#t:callback_opt/0","title":"Hook.callback_opt/0","type":"type"},{"doc":"","ref":"Hook.html#t:callback_opts/0","title":"Hook.callback_opts/0","type":"type"},{"doc":"","ref":"Hook.html#t:count/0","title":"Hook.count/0","type":"type"},{"doc":"","ref":"Hook.html#t:fun_key/0","title":"Hook.fun_key/0","type":"type"},{"doc":"","ref":"Hook.html#t:group/0","title":"Hook.group/0","type":"type"},{"doc":"Examples","ref":"examples.html","title":"Examples","type":"extras"},{"doc":"defmodule Channel do use Hook # 1: import hook/1 macro def events_from_database(), do: # ... def flush() do :ok = events_from_database() |&gt; Enum.map(&amp;encode/1) |&gt; Enum.join() # 2: hook a term |&gt; hook(Mqtt).publish() end def submit(payload), do: # ... end defmodule ChannelTest do test &quot;submitted events are published as single message&quot; do payload_1 = &quot;1&quot; payload_2 = &quot;2&quot; payload = Enum.join([payload_1, payload_2] # 3: define a callback Hook.callback(Mqtt, :publish, fn ^payload -&gt; :ok end) :ok = Channel.submit(event_1) :ok = Channel.submit(event_2) :ok = Channel.flush() # 4: check for unresolved callbacks Hook.assert() end end Exceptions Omitting #3 above would cause: ** (RuntimeError) Hook: failed to resolve a mapping for Mqtt (hook 0.1.0) lib/hook/server.ex:210: Hook.Server.raise_no_mapping_found/1 This is because the hook created at #2 does not Mqtt to another value, rather, it only allows the resolution of a mapping for Mqtt to be delayed until runtime. The callback at #3 is actually what ensures a mapping exists for Mqtt that will result in function calls on it being traced so that assertions may be made. Altering #3 so that the function name :publish was something else, say :publish2, would cause: ** (RuntimeError) Hook: failed to resolve a Mqtt.publish/1 callback for #PID&lt;0.225.0&gt; (hook 0.1.0) lib/hook/callbacks.ex:111: Hook.Callbacks.raise_no_callback_found/4 (hook 0.1.0) lib/hook/server.ex:142: Hook.Server.handle_call/3 We no longer get the previous exception because a mapping does exist for Mqtt, we did define a callback for it, but the callback we defined did not match what the code was calling. Note: This exception helps find bugs in the code you are testing that would cause it to call a function on a hooked module more times than you have defined callbacks for it to resolve. A bug in your code that caused the callback to never be resolved would cause: ** (RuntimeError) Hook: unresolved callbacks for #PID&lt;0.672.0&gt;: {1, Mqtt.publish/1} Either you are defining more callbacks than intended or you&#39;re not calling the function in question as many times as intended. Callbacks Leverage multiple clauses. Hook.callback(System, :cmd, fn &quot;joy&quot;, [&quot;--flag&quot;] -&gt; {&quot;&quot;, 0} &quot;joy&quot;, args -&gt; {&quot;error: --flag required&quot;, 1} end) Define a callback multiple times like this. Hook.callback(System, :cmd, fn cmd, args -&gt; {&quot;joy&quot;, 0} end, count: 2) Or separately. Hook.callback(System, :cmd, fn cmd, args -&gt; {&quot;joy&quot;, 0} end) Hook.callback(System, :cmd, fn cmd, args -&gt; {&quot;joy&quot;, 0} end) Either of those examples could be consumed twice. :count may also be passed :infinity which is a special case. Infinity and beyond Callbacks with an infinity count can be consumed infinitely. Only one infinity count can be defined per module and function at a time. The most recent definition persists. Infinity callbacks are only consumed once all other non-infinity callbacks are exhausted. Assertions Hook.callback(System, :cmd, fn _, _ -&gt; {&quot;joy&quot;, 0} end) Hook.assert() Without the assert line, the above snippet would execute without fail. Adding the assert line causes the following exception. ** (RuntimeError) Hook: unresolved callbacks for #PID&lt;0.672.0&gt;: {1, Mqtt.publish/1}","ref":"examples.html#module-mocks-via-hooks-callbacks-and-asserts","title":"Examples - Module mocks via hooks, callbacks, and asserts","type":"extras"},{"doc":"This is a simple example to demonstrate a common pattern of wiring things together. The App.MqttMock could have a much more complex implementation. defmodule App.MqttMock do require Logger def publish(payload) do Logger.debug(&quot;published \#{inspect payload}&quot;) :ok end end defmodule App.Channel do # 1: import hook/1 macro use Hook def events_from_database(), do: # ... def submit(payload), do: # ... def flush() do :ok = events_from_database() |&gt; Enum.map(&amp;encode/1) |&gt; Enum.join() # 2: hook a term |&gt; hook(Mqtt).publish() end end defmodule ChannelTest do test &quot;flushing multiple events&quot; do Hook.put(Mqtt, App.MqttMock) :ok = Channel.submit(&quot;1&quot;) :ok = Channel.submit(&quot;2&quot;) assert :ok = Channel.flush() end end","ref":"examples.html#module-resolution","title":"Examples - Module resolution","type":"extras"}]